------------------------------------------------------------------------------
--
--  procedure Find_Particles (body)
--
--  The program reads a "Portable Graymap" (PGM) data stream from standard
--  input, and identifies and locates particles with an intensity at or above
--  the threshold intensity.
--
--  Command line arguments:
--    -id            <image id>
--    -threshold     <intensity threshold>
--    -max_particles <maximum particle count>
--
--  Standard input:
--    A plain ASCII PGM file.
--
--  Named file input:
--    (none)
--
--  Standard output:
--    Zero or more lines with the following fields (separated by Latin-1 HT
--    characters):
--      Image id
--      X coordinate
--      Y coordinate
--      Total intensity
--
--  Named file output:
--    (none)
--
------------------------------------------------------------------------------
--  Update information:
--
--  2001.02.10 (Jacob Sparre Andersen)
--    Copied from the soap film data processing procedure Find_Particles
--      (1996.11.18 edition).
--    Reads the image from standard input.
--    Updated the program description.
--
--  (Insert additional update information above this line.)
------------------------------------------------------------------------------
--  Standard packages:

with Ada.Characters.Latin_1;
with Ada.Float_Text_IO;
with Ada.Integer_Text_IO;
with Ada.Text_IO;

------------------------------------------------------------------------------
--  Other packages:

with Generic_Command_Line_Types;
with Generic_Command_Line_Processing;
with Image_Processing;
with PGM;
with UStrings;

------------------------------------------------------------------------------

procedure Find_Particles is

   ---------------------------------------------------------------------------
   --  type Command_Line_Argument_Names:

   type Command_Line_Argument_Names is (ID, Threshold, Max_Particles);

   ---------------------------------------------------------------------------
   --  package Command_Line_Types:

   package Command_Line_Types is new Generic_Command_Line_Types
     (Argument_Names => Command_Line_Argument_Names);

   ---------------------------------------------------------------------------
   --  Command_Line_Help:

   Command_Line_Help : constant Command_Line_Types.Help_Array :=
     (ID            => UStrings.U ("<image id>"),
      Threshold     => UStrings.U ("<intensity threshold>"),
      Max_Particles => UStrings.U ("<maximum particle count>"));

   ---------------------------------------------------------------------------
   --  package Command_Line:

   package Command_Line is new Generic_Command_Line_Processing
     (Command_Line_Types  => Command_Line_Types,
      Obligatory          => (others => True),
      Minimum_Field_Count => (others => 1),
      Maximum_Field_Count => (others => 1),
      Help                => Command_Line_Help);

   ---------------------------------------------------------------------------
   --  package Processing:

   package Processing is new Image_Processing
     (Pixel_Type     => PGM.Grey_16_Bit,
      Image_Handling => PGM.Pixmaps_16_Bit);

   ---------------------------------------------------------------------------
   --  function Maximum_Particle_Count:

   function Maximum_Particle_Count return Processing.Label_Type is

      use Command_Line;
      use Processing;

   begin --  Maximum_Particle_Count
      return Label_Type'Value (Value (Max_Particles, 1));
   end Maximum_Particle_Count;

   ---------------------------------------------------------------------------
   --  Intensity_Threshold:

   subtype Threshold_Range is PGM.Grey_16_Bit range 1 .. PGM.Grey_16_Bit'Last;

   Intensity_Threshold : constant Threshold_Range
     := Threshold_Range (Integer'(Command_Line.Value (Threshold, 1)));

   ---------------------------------------------------------------------------
   --  function Above_Threshold:

   function Above_Threshold (Intensity : in PGM.Grey_16_Bit) return Boolean is

      use type PGM.Grey_16_Bit;

   begin -- Above_Threshold
      return Intensity >= Intensity_Threshold;
   end Above_Threshold;

   pragma Inline (Above_Threshold);

   ---------------------------------------------------------------------------
   --  type Point_Data:

   type Point_Data is
      record
         X, Y      : Integer := 0;
         Intensity : PGM.Grey_16_Bit := 0;
      end record;

   type Point_List is array (Positive range <>) of Point_Data;

   Null_Point : constant Point_Data := (x         => 0,
                                        y         => 0,
                                        Intensity => 0);

   ---------------------------------------------------------------------------
   --  type Particle_Data:

   subtype Label_Type is Processing.Label_Type;

   type Particle_Data is
      record
         X_Sum, Y_Sum  : Natural := 0;
         Intensity_Sum : Natural := 0;
      end record;

   type Particle_List is array (Label_Type range <>) of Particle_Data;
   type Particle_List_Reference is access all Particle_List;

   Null_Particle : constant Particle_Data := (others => 0);

   ---------------------------------------------------------------------------
   --  procedure Append_Point:

   procedure Append_Point (Particle : in out Particle_Data;
                           Pixel    : in     PGM.Grey_16_Bit;
                           X, Y     : in     Integer) is

      use Ada.Text_IO;

   begin -- Append_Point
      Particle := (X_Sum         => Particle.X_Sum + X *  Natural (Pixel),
                   Y_Sum         => Particle.Y_Sum + Y *  Natural (Pixel),
                   Intensity_Sum => Particle.Intensity_Sum + Natural (Pixel));
   exception
      when others =>
         Put_Line
           (Current_Error,
            "Find_Particles.Append_Point: Unexpected exception.");
         raise;
   end Append_Point;

   ---------------------------------------------------------------------------
   -- Process_Image:

   function Process_Image is new Processing.Find_Particles
     (Particle_Type           => Particle_Data,
      Particle_List           => Particle_List,
      Particle_List_Reference => Particle_List_Reference,
      Null_Particle           => Null_Particle,
      Is_A_Particle           => Above_Threshold,
      Append                  => Append_Point);

   ---------------------------------------------------------------------------
   --  procedure Put:
   --
   --  Writes particle data to a file.

   procedure Put (File : in     Ada.Text_IO.File_Type;
                  ID   : in     String;
                  Item : in     Particle_Data) is

      use Ada.Float_Text_IO;
      use Ada.Integer_Text_IO;
      use Ada.Text_IO;

      package Latin_1 renames Ada.Characters.Latin_1;

   begin --  Put
      Put (File => File,
           Item => ID);
      Put (File => File,
           Item => Latin_1.HT);
      Put (File => File,
           Item => Float (Item.X_Sum) / Float (Item.Intensity_Sum),
           Fore => 4,
           Aft  => 4,
           Exp  => 0);
      Put (File => File,
           Item => Latin_1.HT);
      Put (File => File,
           Item => Float (Item.Y_Sum) / Float (Item.Intensity_Sum),
           Fore => 4,
           Aft  => 4,
           Exp  => 0);
      Put (File => File,
           Item => Latin_1.HT);
      Put (File => File,
           Item => Item.Intensity_Sum);
      New_Line (File => File);
   end Put;

   ---------------------------------------------------------------------------

   use Ada.Text_IO;

   Image_ID : constant String := Command_Line.Value (ID, 1);

   Image     : PGM.Pixmaps_16_Bit.Pixmap_Reference;
   Particles : Particle_List_Reference;

begin
   PGM.Load (File => Standard_Input,
             Item => Image);

   Particles := Process_Image (Image, Maximum_Particle_Count);

   for Num in Particles'Range loop
      Put (File => Standard_Output,
           ID   => Image_ID,
           Item => Particles (Num));
   end loop;
exception
   when Processing.Out_Of_Labels =>
      Put_Line
        (File => Current_Error,
         Item => "Find_Particles ran out of labels while processing image " &
                 Image_ID & " before all the particles were identified. " &
                 "Please try again with a higher guess of the maximum " &
                 "number of particles.");
      return;
   when others =>
      Put_Line
        (File => Current_Error,
         Item => "Find_Particles was aborted by an undocumented exception. " &
                 "You may want to yell some curses at the programmer.");
      raise;
end Find_Particles;
