------------------------------------------------------------------------------
--
--  package Image_Processing (body)
--
------------------------------------------------------------------------------
--  Update information:
--
--  1996.04.02 (Jacob Sparre Andersen)
--   Written. Partially based on package Billedbehandling (1995.11.14)
--
--  1996.04.05 (Jacob Sparre Andersen)
--   Inserted missing code in Label_Particles.
--
--  1996.07.26 (Jacob Sparre Andersen)
--    Adjusted formatting to fit the Style Guide style.
--    Reduced the range of type Label by approximately 1/2.
--
--  1996.08.05 (Jacob Sparre Andersen)
--    Inserted additional debugging information and exception reporting in
--      procedure Label_Particles and function Find_Particles.
--
--  1996.11.18 (Jacob Sparre Andersen)
--    Added "Expected_Number_of_Particles" to the argument list for
--      procedure Label_Particles and function Find_Particles.
--
--  1997.09.28-29 (Jacob Sparre Andersen)
--    Switched to access types where apropriate.
--
--  2001.02.10 (Jacob Sparre Andersen)
--    Checking for available labels in procedure Label_Particles.Choose_Label.
--
--  (Insert additional update information above this line.)
------------------------------------------------------------------------------
--  Standard packages:

with Ada.Text_IO;

------------------------------------------------------------------------------

with Debugged;

package body Image_Processing is

   use Debugged;

   ---------------------------------------------------------------------------
   --  procedure Label_Particles:
   --
   --  The image and the label map should have the same coordinate ranges.

   procedure Label_Particles
     (Image                        : in     Image_Reference;
      Labels                       : in     Label_Map_Reference;
      Number_of_Particles          :    out Integer;
      Expected_Number_of_Particles : in     Label_Type := Label_Type'Last) is

      ---------------------------------------------------------------------
      --  Image size:

      Width  : constant Natural := Image.all'Length (1);
      Height : constant Natural := Image.all'Length (2);

      ---------------------------------------------------------------------

      Maximum_Number_Of_Particles : constant Label_Type :=
        Label_Type'Min (Label_Type (Width * Height / 2 + 1),
                        Expected_Number_of_Particles);

      ---------------------------------------------------------------------
      --  subtype Label:

      subtype Label is Label_Type range 0 .. Maximum_Number_Of_Particles;

      ---------------------------------------------------------------------
      --  type Particle_Alias_List:

      type Particle_Alias_List is array (Label range <>) of Label;

      type Particle_Alias_List_Reference is access all Particle_Alias_List;

      ---------------------------------------------------------------------
      --  Coordinate subtypes:

      subtype X_Range is Integer range Image.all'Range (1);
      subtype Y_Range is Integer range Image.all'Range (2);

      ---------------------------------------------------------------------     

      Last_Used_Label : Label := 0;
      Alias_List      : Particle_Alias_List_Reference;

      ---------------------------------------------------------------------    
      --  procedure Choose_Label:

      procedure Choose_Label (X : in     X_Range;
                              Y : in     Y_Range) is

      begin -- Choose_Label
         if Labels (X, Y) = 0 then
            if X = X_Range'First then
               if (y > Y_Range'First) and then (Labels (x, y - 1) > 0) then
                  Labels (x, y) := Labels (x, y - 1);
               elsif Last_Used_Label < Label'Last then
                  Last_Used_Label := Last_Used_Label + 1;
                  Labels (x, y) := Last_Used_Label;
                  Alias_List (Last_Used_Label) := Last_Used_Label;
               else
                  raise Out_Of_Labels;
               end if;
            elsif y = Y_Range'First then
               if (x > X_Range'First) and then
                  (Labels (x - 1, y) > 0) then
                  Labels (x, y) := Labels (x - 1, y);
               elsif Last_Used_Label < Label'Last then
                  Last_Used_Label := Last_Used_Label + 1;
                  Labels (x, y) := Last_Used_Label;
                  Alias_List (Last_Used_Label) := Last_Used_Label;
               else
                  raise Out_Of_Labels;
               end if;
            else
               if (Labels (x - 1, y) > 0) and
                  (Labels (x, y - 1) > 0) then
                  Labels (x, y) := Labels (x - 1, y);
                  Alias_List (Labels (x, y - 1)) := Labels (x - 1, y);
               elsif Labels (x - 1, y) > 0 then
                  Labels (x, y) := Labels (x - 1, y);
               elsif Labels (x, y - 1) > 0 then
                  Labels (x, y) := Labels (x, y - 1);
               elsif Last_Used_Label < Label'Last then
                  Last_Used_Label := Last_Used_Label + 1;
                  Labels (x, y) := Last_Used_Label;
                  Alias_List (Last_Used_Label) := Last_Used_Label;
               else
                  raise Out_Of_Labels;
               end if;
            end if;
         end if;
      end Choose_Label;

   begin --  Label_Particles
      Message_Line ("Image_Processing.Label_Particles:");

      Message_Line ("  Image size: " & Width'Img & " * " & Height'Img);
      Message_Line ("  " & X_Range'First'Img & " = " & Image.all'First (1)'Img);

      Message_Line ("  Max number of labels: " & Label'Image (Label'Last));

      Message ("  Allocating particle alias list ... ");
      Alias_List := new Particle_Alias_List (1 .. Label'Last);
      Message_Line ("Done!");

      Labels.all := (others => (others => 0));
      Alias_List.all := (others => 0);

      Message_Line ("  loop Label_Particles");

   Label_Particles:
      for X in X_Range loop
         for Y in Y_Range loop
            if Is_A_Particle (Image (X, Y)) then
--Message (" [ (" & X'Img & ", " & Y'Img & ") is a particle ] ");
               Choose_Label (X, Y);
            end if;
         end loop;
      end loop Label_Particles;

      Message_Line ("  end loop Label_Particles");

   Count_Unique_Labels:
      declare

         Counted_Unique_Labels : Label := 0;

      begin
         Message_Line ("  Counting unique labels...");

         for Label in 1 .. Last_Used_Label loop
            if Alias_List (Label) = Label then
               Counted_Unique_Labels := Counted_Unique_Labels + 1;
            end if;
         end loop;

         Message_Line ("  Unique labels counted.");

         declare

            Number_Of_Unique_Labels : constant Label := Counted_Unique_Labels;

            type Unique_Label is new Label range 0 .. Number_Of_Unique_Labels;

            Alias_For_Unique_Label : array (Label) of Unique_Label := 
             (Others => 0) ;

            --------------------------------------------------------------
            -- Get_Unique_Label:

            -- Those labels that points to themselves should be assigned
            -- an unique label before Get_Unique_Label is called.
            --
            -- The function calls itself recursively until it gets to
            -- a label that has been assigned an unique label.

            function Get_Unique_Label (L : in Label) return Unique_Label is

            begin -- Get_Unique_Label
               if (L = 0) or else (Alias_List (L) = 0) then
                  return 0;
               else
                  Message_Line ("    Get_Unique_Label (" & Label'Image (L) & ")");

                  if Alias_For_Unique_Label (L) = 0 then
                     Alias_For_Unique_Label (L) := Get_Unique_Label (Alias_List (L));
                  end if;
                  return Alias_For_Unique_Label (L);
               end if;
            end Get_Unique_Label;

            --------------------------------------------------------------

         begin
            Message_Line ("  Making alias list for the unique labels...");

            Counted_Unique_Labels := 0;
            for Label in Alias_List'Range loop
               if Alias_List (Label) = Label then
                  Counted_Unique_Labels := Counted_Unique_Labels + 1;
                  Alias_For_Unique_Label (Label) :=
                    Unique_Label (Counted_Unique_Labels);
               end if;
            end loop;

            Message_Line ("  Done making alias list for the unique labels.");

            Message_Line ("  Relabeling points...");

         Relabel_Particles:
            for X in X_Range loop
               for Y in Y_Range loop
                  Labels (X, Y) := Label_Type (Get_Unique_Label (Labels (X, Y)));
               end loop;
            end loop Relabel_Particles;

            Message_Line ("  Relabeled points.");
            Message_Line ("  Number of unique labels = " &
                            Label'Image (Number_Of_Unique_Labels));

            Number_of_Particles := Integer (Number_Of_Unique_Labels);
         end;
      end Count_Unique_Labels;
   exception
      when Out_Of_Labels =>
         raise;
      when others =>
         Ada.Text_IO.Put_Line
           (File => Ada.Text_IO.Current_Error,
            Item => "An undocumented exception was raised in procedure " &
                    "Image_Processing.Label_Particles. You may want to " &
                    "yell at the programmer.");
         raise;
   end Label_Particles;

   ---------------------------------------------------------------------------
   --  function Find_Particles:

   function Find_Particles
     (Image                        : in     Image_Reference;
      Expected_Number_of_Particles : in     Label_Type := Label_Type'Last) 
     return Particle_List_Reference is

      ---------------------------------------------------------------------
      --  procedure Find_Labels:

      procedure Find_Labels is new Label_Particles (Is_A_Particle);

      ---------------------------------------------------------------------

      Labels         : Label_Map_Reference;
      Particle_Count : Integer;
      Particles      : Particle_List_Reference;

   begin --  Find_Particles
      Message_Line ("Image_Processing.Find_Particles: begin ...");

      Message ("  Allocating label map ... ");
      Labels := new Label_Map (Image'Range (1), Image'Range (2));
      Message_Line ("Done!");

      Message ("  Finding labels ... ");
      Find_Labels
        (Image                        => Image,
         Labels                       => Labels,
         Number_Of_Particles          => Particle_Count,
         Expected_Number_of_Particles => Expected_Number_of_Particles);
      Message_Line ("Done!");

      Message ("  Allocating particle list ... ");
      Particles :=
        new Particle_List'(1 .. Label_Type (Particle_Count) => Null_Particle);
      Message_Line ("Done!");

      for x in Image'Range (1) loop
         for y in Image'Range (2) loop
            if Labels (x, y) > 0 then

            Append_Particle:
               begin
                  Append (Particles (Labels (x, y)),
                          Image (x, y),
                          x, y);
               exception
                  when others =>
                     Ada.Text_IO.Put_Line
                       (Ada.Text_IO.Current_Error,
                        "Image_Processing.Find_Particles.Append_Particle:" &
                          " Unexpected exception. Probably caused by an " &
                          "error in procedure Append.");
                     raise;
               end Append_Particle;
            end if;
         end loop;
      end loop;

      Message_Line ("Image_Processing.Find_Particles: Ready to return.");

      return Particles;
   exception
      when Out_Of_Labels =>
         raise;
      when others =>
         Ada.Text_IO.Put_Line
           (File => Ada.Text_IO.Current_Error,
            Item => "An undocumented exception was raised in procedure " &
                    "Image_Processing.Find_Particles. You may want to " &
                    "yell at the programmer.");
         raise;
   end Find_Particles;

   -------------------------------------------------------------------------

end Image_Processing;
